class _BaseException(Exception):
    def __init__(self, *a, **kw):
        handle = kw.pop('handle', None)
        super(Exception, self).__init__(*a, **kw)
        self.handle = handle


class WorkFailed(_BaseException):
    pass


class WorkException(_BaseException):
    pass



import re
import asyncio
import logging
from collections import deque
from io import StringIO

logger = logging.getLogger(__name__)


class Admin(asyncio.Protocol):
    def __init__(self, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.transport = None
        self.waiters = deque()
        self.buffer = b''
        self.linesep = b'\n'
        self.structure_sep = self.linesep + b'.' + self.linesep
        self.telnet_sep = b'\r\n'

    def connection_made(self, transport):
        self.transport = transport

    def connection_lost(self, exc):
        self.transport = None

    def data_received(self, data):
        self.buffer += data
        while self.buffer:
            try:
                pair = self.waiters[0]
                f, eol = pair
            except IndexError:
                logger.warning('Unexpected msg %s', self.buffer)
                break

            buffer = self.buffer
            index = buffer.find(eol)
            if index < 0:
                break

            msg = buffer[:index]
            self.buffer = buffer[index+len(eol):]
            self.waiters.popleft()
            f.set_result(msg.decode('utf-8'))

    async def workers(self):
        resp = await self.send_and_wait_resp(b'workers', self.structure_sep)
        sio = StringIO(resp)
        workers = []
        for line in sio:
            client, funcs = line.split(':')
            try:
                fd, ip, name = re.findall(r'(\S+)', client)
                funcs = re.findall(r'(\S+)', funcs)
            except ValueError:
                logger.warning('Malformed worker line %s', line)
                continue

            worker = dict(fd=int(fd), ip_address=ip, name=name, functions=funcs)
            workers.append(worker)
        return workers

    async def status(self):
        resp = await self.send_and_wait_resp(b'status', self.structure_sep)
        sio = StringIO(resp)
        status = []
        for line in sio:
            func, *stats = re.findall(r'(\S+)', line)
            total, running, available = map(int, stats)
            entry = dict(function=func, total=total, running=running, available_workers=available)
            status.append(entry)
        return status

    async def maxqueue(self, function, size=None):
        try:
            function = function.encode('ascii')
        except AttributeError:
            pass
        data = [b'maxqueue', function]
        if size is not None:
            try:
                size = str(int(size)).encode('ascii')
                data.append(size)
            except (TypeError, ValueError):
                pass
            try:
                if len(size) != 3:
                    raise RuntimeError
                size = [str(int(s)).encode('ascii') for s in size]
                data.extend(size)
            except:
                raise RuntimeError(
                    'Unsupported size parameter {}, it must be either int or list of 3 values')
        return await self.send_and_wait_resp(b' '.join(data), self.telnet_sep)

    async def shutdown(self, graceful=False):
        data = [b'shutdown']
        if graceful:
            data.append(b'graceful')
        return await self.send_and_wait_resp(b' '.join(data), self.telnet_sep)

    async def version(self):
        return await self.send_and_wait_resp(b'version', self.linesep)

    async def verbose(self):
        return await self.send_and_wait_resp(b'verbose', self.linesep)

    def send_and_wait_resp(self, cmd, expected_eol):
        self.transport.write(cmd + self.linesep)
        f = self.loop.create_future()
        pair = (f, expected_eol)
        self.waiters.append(pair)
        return f

    def disconnect(self):
        try:
            self.transport.close()
        except AttributeError:
            pass

    close = disconnect



import asyncio
import logging
import uuid
import random
from functools import partial
from aiogear.packet import Type
from aiogear.mixin import GearmanProtocolMixin

logger = logging.getLogger(__name__)


class Client(GearmanProtocolMixin, asyncio.Protocol):
    def __init__(self, loop=None):
        super(Client, self).__init__(loop=loop)
        self.transport = None

        self.submit_job = partial(self._submit_job, Type.SUBMIT_JOB)
        self.submit_job_bg = partial(self._submit_job, Type.SUBMIT_JOB_BG)
        self.submit_job_high = partial(self._submit_job, Type.SUBMIT_JOB_HIGH)
        self.submit_job_high_bg = partial(self._submit_job, Type.SUBMIT_JOB_HIGH_BG)
        self.submit_job_low = partial(self._submit_job, Type.SUBMIT_JOB_LOW)
        self.submit_job_low_bg = partial(self._submit_job, Type.SUBMIT_JOB_LOW_BG)

        self.handles = {}
        self._closing = None

    def __del__(self):
        try:
            self.disconnect()
        except AttributeError:
            pass

    @staticmethod
    def uuid():
        # 0x00 is used as delimiter in gearman protocol
        # replace it with random printable character.
        replacement = chr(random.randint(32, 126)).encode('ascii')
        return uuid.uuid4().bytes.replace(b'\0', replacement)

    def connection_made(self, transport):
        self.transport = transport

    def connection_lost(self, exc):
        self.transport = None
        if self._closing:
            self._closing.set_result(exc)

    async def _submit_job(self, packet, name, *args, **kwargs):
        uuid = kwargs.pop('uuid', None)
        if uuid is None:
            uuid = self.uuid()
        self.send(packet, name, uuid, *args)
        job_created = await self.wait_for(Type.JOB_CREATED)
        f = self.wait_for(Type.WORK_COMPLETE, Type.WORK_FAIL, Type.WORK_EXCEPTION)
        handle = job_created.handle
        self.handles[handle] = f
        f.add_done_callback(lambda _: self.handles.pop(handle))
        return job_created

    def submit_job_sched(self, name, dt, *args, **kwargs):
        sched_args = [str(int(x)) for x in dt.strftime('%M %H %d %m %w').split()]
        return self._submit_job(Type.SUBMIT_JOB_SCHED, name, *(sched_args + list(args)), **kwargs)

    def get_status(self, handle):
        self.send(Type.GET_STATUS, handle)
        return self.wait_for(Type.STATUS_RES)

    def get_status_unique(self, uuid):
        self.send(Type.STATUS_RES_UNIQUE, uuid)
        return self.wait_for(Type.STATUS_RES_UNIQUE)

    def option_req(self, option):
        self.send(Type.OPTION_REQ, option)
        return self.wait_for(Type.OPTION_RES, Type.ERROR)

    def wait_job(self, handle):
        f = self.handles.get(handle)
        if not f:
            raise RuntimeError('Unable to find handle {} in handles {}'.format(handle, self.handles))
        return f

    def disconnect(self):
        self.transport.close()
        f = self._closing = self.loop.create_future()
        return f

    close = disconnect



from aiogear.worker import Worker
from aiogear.client import Client
from aiogear.admin import Admin
from aiogear.packet import Type as PacketType


__all__ = ['Worker', 'Client', 'Admin', 'PacketType']



from enum import Enum


class Type(Enum):
    CAN_DO = 1
    CANT_DO = 2
    RESET_ABILITIES = 3
    PRE_SLEEP = 4
    # Unused
    NOOP = 6
    SUBMIT_JOB = 7
    JOB_CREATED = 8
    GRAB_JOB = 9
    NO_JOB = 10
    JOB_ASSIGN = 11
    WORK_STATUS = 12
    WORK_COMPLETE = 13
    WORK_FAIL = 14
    GET_STATUS = 15
    ECHO_REQ = 16
    ECHO_RES = 17
    SUBMIT_JOB_BG = 18
    ERROR = 19
    STATUS_RES = 20
    SUBMIT_JOB_HIGH = 21
    SET_CLIENT_ID = 22
    CAN_DO_TIMEOUT = 23
    ALL_YOURS = 24
    WORK_EXCEPTION = 25
    OPTION_REQ = 26
    OPTION_RES = 27
    WORK_DATA = 28
    WORK_WARNING = 29
    GRAB_JOB_UNIQ = 30
    JOB_ASSIGN_UNIQ = 31
    SUBMIT_JOB_HIGH_BG = 32
    SUBMIT_JOB_LOW = 33
    SUBMIT_JOB_LOW_BG = 34
    SUBMIT_JOB_SCHED = 35
    SUBMIT_JOB_EPOCH = 36
    SUBMIT_REDUCE_JOB = 37
    SUBMIT_REDUCE_JOB_BACKGROUND = 38
    GRAB_JOB_ALL = 39
    JOB_ASSIGN_ALL = 40
    GET_STATUS_UNIQUE = 41
    STATUS_RES_UNIQUE = 42



from setuptools import setup

setup(
    name='aiogear',
    version='0.2.1',
    author='Sinan Nalkaya',
    author_email='sardok@gmail.com',
    url='https://github.com/sardok/aiogear',
    description='Asynchronous gearman protocol based on asyncio',
    packages=['aiogear'],
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)
